package dareka;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Collections;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * 
 * Abstraction of cache handling.
 * 
 * Naming specification:
 * 
 * A cache is identified by cacheId which consists of letters without '_'.
 * 
 * A cache may have a description which consists of letters being available for
 * filename.
 * 
 * Mapping of cache names and filename: cacheId + ('_' + description)? + '.flv'
 * 
 */
public class Cache {
    private String cacheId;
    private static File cacheDir = new File("cache");
    private static final Pattern CACHE_FILE_PATTERN = Pattern
            .compile("^([^_]+?)(?:_.*)?\\.flv$");
    private static final Pattern NUMBER_CHARACTER_REFERENCE_PATTERN = Pattern
            .compile("&#(\\d+);");
    private static ConcurrentHashMap<String, File> id2File = new ConcurrentHashMap<String, File>();
    private File cacheFile;
    private File tmpFile;

    public static void init() {
        cacheDir.mkdir();

        searchCachesOnADirectory(cacheDir);
    }

    private static void searchCachesOnADirectory(File dir) {
        File[] cacheFiles = dir.listFiles();

        for (File file : cacheFiles) {
            if (file.isDirectory()) {
                searchCachesOnADirectory(file);
            } else if (file.isFile()) {
                String id = getIdFromFilename(file.getName());
                if (!id.equals("")) {
                    id2File.put(id, file);

                    Logger
                            .debug("cache found: " + id + " => "
                                    + file.getPath());
                }
            }
        }
    }

    static String getIdFromFilename(String filename) {
        Matcher m = CACHE_FILE_PATTERN.matcher(filename);
        if (m.find()) {
            return m.group(1);
        } else {
            return "";
        }
    }

    public static Map<String, File> getId2File() {
        return Collections.unmodifiableMap(id2File);
    }

    public Cache(String cacheId) {
        this.cacheId = cacheId;

        cacheFile = id2File.get(cacheId);
        if (cacheFile == null) {
            cacheFile = new File(cacheDir, cacheId + ".flv");
        }
    }

    public Cache(String cacheId, String desc) {
        this.cacheId = cacheId;

        cacheFile = id2File.get(cacheId);
        if (cacheFile == null) {
            cacheFile = getDescribedCacheFile(cacheId, desc);
        }
    }

    static File getDescribedCacheFile(String cacheId, String desc) {
        StringBuffer decodedDesc = new StringBuffer(desc.length());

        Matcher m = NUMBER_CHARACTER_REFERENCE_PATTERN.matcher(desc);

        while (m.find()) {
            m.appendReplacement(decodedDesc, Character.toString((char) Integer
                    .parseInt(m.group(1))));
        }
        m.appendTail(decodedDesc);

        String sanitizedDesc = getSanitizedDescription(decodedDesc.toString());

        return new File(cacheDir, cacheId + '_' + sanitizedDesc + ".flv");
    }

    static String getSanitizedDescription(String decodedDesc) {
        StringBuilder sanitizedDesc = new StringBuilder(decodedDesc.length());

        for (int i = 0; i < decodedDesc.length(); i++) {
            char c = decodedDesc.charAt(i);
            if (c == '\\' || c == '/' || c == ':' || c == ',' || c == ';'
                    || c == '*' || c == '?' || c == '"' || c == '<' || c == '>'
                    || c == '|') {
                sanitizedDesc.append('-');
            } else {
                sanitizedDesc.append(c);
            }
        }
        return sanitizedDesc.toString();
    }

    private File getCacheTmpFile() throws IOException {
        if (tmpFile == null) {
            tmpFile = File.createTempFile(cacheId + ".flv-", ".tmp", cacheDir);
        }

        return tmpFile;
    }

    public boolean exists() {
        return cacheFile.exists();
    }

    public InputStream getInputStream() throws IOException {
        cacheFile.setLastModified(System.currentTimeMillis());
        return new FileInputStream(cacheFile);
    }

    public OutputStream getTmpOutputStream() throws IOException {
        File cacheTmpFile = getCacheTmpFile();
        return new FileOutputStream(cacheTmpFile);
    }

    public void deleteTmp() throws IOException {
        File cacheTmpFile = getCacheTmpFile();
        cacheTmpFile.delete();
    }

    public void store() throws IOException {
        File cacheTmpFile = getCacheTmpFile();

        if (cacheTmpFile.renameTo(cacheFile)) {
            id2File.put(cacheId, cacheFile);
        } else {
            cacheTmpFile.delete();
        }
    }

    public long tmpLength() throws IOException {
        return getCacheTmpFile().length();
    }

    public long length() {
        return cacheFile.length();
    }

    public static void cleanup() {
        File[] tmpFiles = cacheDir.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                if (name.endsWith(".tmp")) {
                    return true;
                }
                return false;
            }
        });

        for (File file : tmpFiles) {
            // directory is also target, but removed only if it's empty.
            file.delete();
        }
    }

    public static long size() {
        long sum = 0;
        
        for (File file : id2File.values()) {
            sum += file.length();
        }
        
        return sum;
    }

    public String getId() {
        return cacheId;
    }

    public String getCacheFileName() {
        return cacheFile.getName();
    }

    public File getCacheFile() {
        return cacheFile;
    }
}
